\def\SeminarColorVersion{y}
\def\SeminarPaperVersion{n}

\documentclass{seminar}

\usepackage{sem-dem}            % Small utility macros for these demo files
\usepackage{Haskell2LaTeX}
\usepackage{calc}
\usepackage{example}

\def\hkeyword#1{\hbox{\textbf{#1}}}
\def\hvarid#1{\hbox{\textit{#1}}}
\def\htycon#1{\hbox{\textsf{#1}}}
\def\filename#1{\hbox{\texttt{#1}}}
\def\program#1{\hbox{\texttt{#1}}}
\def\environment#1{\hbox{\texttt{#1}}}
\def\latexpackage#1{\hbox{\texttt{#1}}}

% Identification of the file (varying parameters)
\def\FileDate{October 2002}
\def\FileVersion{1.0}
\HyperSetUp

% We must redefine the header for this introduction file
\SeminarHeader{\LARGE\HLe{\FileInfoA}}

\VerbatimFootnotes              % To allow veratim material in footnotes

% To print the notes in the paper version only
\ifx\SeminarPaperVersion\AnswerYes
\else
  \slidesonlytrue
\fi

\begin{document}

%\SeminarFirstSlide
\SeminarSlideFrameBackground{SeminarBackgroundSolid}
\begin{slide}
  \ifx\SeminarPaperVersion\AnswerYes
  \else
    \vspace*{\stretch{1}} % Due to the special annotation for full screen mode
  \fi

  \begin{center}
    \psframebox[fillstyle=solid,fillcolor=red,framearc=0.2]{%
      \Large\bfseries\color{yellow}
      \begin{tabular}{c}
        An introduction to\\
        {\Huge\FileInfoA}
      \end{tabular}}

    \vspace{1cm}
    \psframebox[fillstyle=solid,fillcolor=LightBlue,
                framesep=0.2,framearc=0.4]{%
      \bfseries
      \begin{tabular}{c}
        \FileAuthor\\
        \FileDate
      \end{tabular}}
  \end{center}
  
  \ifx\SeminarPaperVersion\AnswerYes
  \else
    \vspace*{\stretch{1}} % Due to the special annotation for full screen mode
  \fi
\end{slide}

%\SeminarListOfSlides

% Background
\ifx\SeminarColorVersion\AnswerYes
  \renewcommand{\SeminarBackgroundGradientColorBegin}{white}
  \renewcommand{\SeminarBackgroundGradientColorEnd}{Lavender}
  \SeminarSlideFrameBackground{SeminarBackgroundGradient}
\fi

\begin{slide}
\vspace*{\fill}
\slideheading{Literate programming}

I believe that the time is ripe for significantly better documentation
of programs, and that we can best achieve this by considering programs
to be works of literature. Hence, my title: ``Literate Programming.''

Let us change our traditional attitude to the construction of programs:
Instead of imagining that our main task is to instruct a computer what
to do, let us concentrate rather on explaining to human beings what we
want a computer to do.

\hspace{5em}-- Donald E. Knuth, ``Literate Programming'', 1984
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begingroup
\renewcommand{\slidestretch}{0.86}% To compress the lines
\begin{slide}
  \vspace*{\fill}
  \slideheading{Objective}

  Transform this:

\begin{verbatim}
  \begin{code}
  qsort :: [a] -> [a]
  qsort []     = []
  qsort (x:xs) = qsort xs_lt_x ++ [x] ++ qsort xs_greq_x
      where xs_lt_x   = [y | y <- xs, y < x]
            xs_greq_x = [y | y <- xs, y >= x]
  \end{code} 
\end{verbatim}

\vspace{0.6em}
into this:
\vspace{-0.6em}
\hspace*{1.6em}\parbox{\hsize-1.6em}{
\begin{code}
qsort :: [a] -> [a]
qsort []     = []
qsort (x:xs) = qsort xs_lt_x ++ [x] ++ qsort xs_greq_x         
    where xs_lt_x   = [y | y <- xs, y < x]                      
          xs_greq_x = [y | y <- xs, y >= x]
\end{code}
}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}
\endgroup

\begin{slide}
\vspace*{\fill}
  \slideheading{Haskell}

  Haskell is...

  \begin{dinglist}[red]{\DingListSymbolA}
    \item the language to be pretty-printed
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item The source format has some support for literate programming
    \end{dinglist}

    \item the language the pretty-printer will be written in
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item Well suited to parsing and tree manipulation
      \item Could you seriously consider anything else?
    \end{dinglist}
  \end{dinglist}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{{\TeX} and {\LaTeX}}

  \begin{dinglist}[red]{\DingListSymbolA}
    \item Literate Haskell makes targetting the {\TeX} family slightly easier
\begin{verbatim}
  \usepackage{verbatim}
  \newenvironment{code}{\verbatim}{\endverbatim}
\end{verbatim}
    \item The {\TeX} family have the mathematical symbols commonly used
          in typesetting Haskell, e.g. $\rightarrow$, $\ge$
    \item Using {\LaTeX} allows you to write your documentation in exactly
          the same way as you (probably!) write your papers etc.
  \end{dinglist}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Parsing}

  Haskell is...

  \begin{dinglist}[red]{\DingListSymbolA}
    \item Much of the art and science of parsing dates from the 70s and 80s
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item Lexical analysis then syntactic analysis
      \item Top-down vs botom-up
    \end{dinglist}
    \item Parser combinators provide a nice way of writing top-down
          parsers in HOT languages
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item Take a list of symbols and return a value
      \item Provide primitive parsers, e.g. Fail, Succeed, Match
      \item Provide a number of combinators, e.g. Choice, Sequential
            Composition
    \end{dinglist}
  \end{dinglist}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Existing pretty-printers}

  \begin{dinglist}[red]{\DingListSymbolA}
    \item Haskell Style for LaTeX2e
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item For manual typesetting
    \end{dinglist}
    \item \texttt{haskell.sty}
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item Easily confused, e.g. \verb!<=>! pretty-prints as $\le>$
    \end{dinglist}
    \item $\lambda$\TeX
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item Forces certain styles on you or indents incorrectly
    \end{dinglist}
    \item \texttt{lhs2tex}
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item The most impressive, but still limited by only lexing the input
    \end{dinglist}
    \item Related programs
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item \texttt{detex}, \texttt{enscript}, HDoc, HaskellDoc,
            The Haskell Module Browser
    \end{dinglist}
  \end{dinglist}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Haskell2{\LaTeX} Structure}

  \begin{dinglist}[red]{\DingListSymbolA}
    \item Parsing Haskell
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item Unlitting
      \item Lexing
      \item Layout rule
      \item Fixity annotation
      \item Syntax analysing
    \end{dinglist}
    \item Pretty-printing the abstract syntax tree
  \end{dinglist}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Unlitting}

  \begin{dinglist}[red]{\DingListSymbolA}
  \item\mbox{}\\
  \vspace{-37.6pt}
\begin{code}
unlit :: Bool -> String -> [Either String (String, [(Char, Position)], Position)]
\end{code}
    \item \verb!.hs! files:
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item Do nothing!
    \end{dinglist}
    \item \verb!.lhs! files:
    \begin{dinglist}[green]{\DingListSymbolB}\setlength{\labelwidth}{4mm}
      \item Extract the contents of \verb!\begin{code}!\dots\verb!\end{code}!
            blocks
      \item Take all lines beginning with \verb!>! and replace the leading
            \verb!>! with a space
      \item Keep the position info!
    \end{dinglist}
  \end{dinglist}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Parser Combinator library types}

  \begin{dinglist}[red]{\DingListSymbolA}
  \item\mbox{}\\
  \vspace{-37.6pt}
\begin{code}
type Parser a = String -> Maybe (a, Int, String)
\end{code}
  \item\mbox{}\\
  \vspace{-37.6pt}
\begin{code}
type Parser a b = [a] -> Maybe (b, Int, [a])
\end{code}
  \item\mbox{}\\
  \vspace{-37.6pt}
\begin{code}
type Parser a b = [(a, Position)] -> Parsed a b
data Parsed a b = Success b Int [(a, Position)]
                | Failure Position
\end{code}
  \end{dinglist}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begingroup
\renewcommand{\slidestretch}{0.46}% To compress the lines
\begin{slide}
\vspace*{\fill}
  \slideheading{Parser Combinator library}

  \begin{dinglist}[red]{\DingListSymbolA}
  \item Primitives:
\begin{code}
pPred :: (a -> Bool) -> Parser a a
pFail :: Parser a b
pSucceed :: b -> Parser a b
(<|>) :: Parser a b -> Parser a b -> Parser a b
(<*>) :: Parser a (b -> c) -> Parser a b -> Parser a c
(<!>) :: Parser a b -> Parser a c -> Parser a b
\end{code}

  \item Derived:
\begin{code}
(<*) :: Parser a b -> Parser a c -> Parser a b
(<$>) :: (b -> c) -> Parser a b -> Parser a c
(<$) :: b -> Parser a c -> Parser a b
\end{code}
  \end{dinglist}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begingroup
\renewcommand{\slidestretch}{0.46}% To compress the lines
\begin{slide}
\vspace*{\fill}
  \slideheading{Parser Combinator library (cont.)}

  \begin{dinglist}[red]{\DingListSymbolA}
  \item Larger parsers:
\begin{code}
pSym :: Eq a => a -> Parser a a
pSyms :: Eq a => [a] -> Parser a [a]
opt :: Parser a b -> b -> Parser a b
pPredHolds :: (a -> Bool) -> Parser a Bool
pMaybe :: Parser a b -> Parser a (Maybe b)
pList :: Int -> Parser a b -> Parser a [b]
pSList :: Bool -> Int -> Parser a b -> Parser a c -> Parser a ([Either b c])
\end{code}

  \item Position handling:
\begin{code}
keep_pos :: Parser a b -> Parser a (b, Position)
(<&>) :: (b -> c) -> Parser a b -> Parser a (c, Position)
(<&) :: b -> Parser a c -> Parser a (b, Position)
\end{code}
  \end{dinglist}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}
\endgroup

\begin{slide}
\vspace*{\fill}
  \slideheading{Derived parsers implementation}

\begin{code}
opt :: Parser a b -> b -> Parser a b
p `opt` v = p <|> if accepts_empty p then pFail else pSucceed v

pList :: Int -> Parser a b -> Parser a [b]
pList _ p | accepts_empty p = error "Tried to make a list of empties"
pList 0 p = pList 1 p `opt` []
pList (i+1) p = (:) <$> p <*> pList i p
\end{code}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Problem with position detection}

\begin{example}
Hugs session for:
/usr/share/hugs98/lib/Prelude.hs
Position.lhs
PCbase.lhs
PC.lhs
PC> \hbox{\textbf{id <$ pSym '\{' <*> pList 0 (pSyms "abcd") <* pSym '\}' $ posify "\{abcdabcdabcd\}"}}
Success ["abcd","abcd","abcd"] 14 []
PC> \hbox{\textbf{id <$ pSym '\{' <*> pList 0 (pSyms "abcd") <* pSym '\}' $ posify "\{abcdabcdabd\}"}}
Failure (Position 1 10)
\end{example}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Solution: Fix the datastructure}

\begin{code}
data Parsed a b = Success (b, Int, [(a, Position)], Position)
                | Failure Position
type Parser a b = [(a, Position)] -> Parsed a b
\end{code}

\begin{code}
pFail :: Parser a b
pFail = \inp -> Failure (get_pos inp)

pSucceed :: b -> Parser a b
pSucceed v = \inp -> Success (v, 0, inp, get_pos inp)

pPred :: (a -> Bool) -> Parser a a
pPred f = \inp -> case inp of
                      [] -> Failure end_of_file
                      ((t, p):inp') -> if f t
                                       then Success (t, 1, inp', get_pos inp')
                                       else Failure p
\end{code}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
\begin{code}
(<|>) :: (Show a) => Parser a b -> Parser a b -> Parser a b
p <|> q
 = \inp -> case (p inp, q inp) of
               (Success s1, Success s2) -> case num s1 `compare` num s2 of
                                               GT -> succeed s1 (pos s2)
                                               LT -> succeed s2 (pos s1)
                                               EQ -> Failure (get_pos inp)
               (Success s1, Failure pos2) -> succeed s1 pos2
               (Failure pos1, Success s2) -> succeed s2 pos1
               (Failure pos1, Failure pos2) -> Failure $! fp
                   where fp = furthest_pos pos1 pos2
    where succeed (v, n, inp, pos) pos' = fp `seq` Success (v, n, inp, fp)
              where fp = furthest_pos pos pos'

(<*>) :: Parser a (b -> c) -> Parser a b -> Parser a c
p1 <*> p2 = \inp -> case p1 inp of
                        Failure pos1 -> Failure pos1
                        Success s1 ->
                            case p2 (rinp s1) of
                                Failure pos2 -> Failure $! fp
                                    where fp = furthest_pos (pos s1) pos2
                                Success s2 -> Success (seq_comp s1 s2)
    where seq_comp (v1, n1, _, pos1) (v2, n2, r, pos2) = fp `seq` (v1 v2, n1 + n2, r, fp)
              where fp = furthest_pos pos1 pos2
\end{code}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Optimisation 1}

\begin{code}
(<|) :: Parser a b -> Parser a b -> Parser a b
p1 <| p2 = \inp -> case (p1 inp, p2 inp) of
                       (Success r1, _) -> Success r1
                       (Failure pos1, Success (v2, n2, inp2, pos2)) ->
                               fp `seq` Success (v2, n2, inp2, fp)
                                   where fp = furthest_pos pos1 pos2
                       (Failure pos1, Failure pos2) ->
                               Failure $! fp
                                   where fp = furthest_pos pos1 pos2

(|>) :: Parser a b -> Parser a b -> Parser a b
p1 |> p2 = p2 <| p1
\end{code}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Optimisation 2}

\begin{code}
type SParsed a b = (b, Int, [(a, Position)], Position)
data Parsed a b = Success (SParsed a b)
                | Failure Position
type SParser a b = [(a, Position)] -> SParsed a b
type Parser a b = [(a, Position)] -> Parsed a b

pList :: Int -> Parser a b -> Parser a [b]
pList 0 p = pList0 p
pList (i+1) p = (:) <$> p <*> pList i p

pList0 :: Parser a b -> Parser a [b]
pList0 p = \inp -> Success (pListS p inp)
pListS :: Parser a b -> SParser a [b]
pListS p = \inp -> case p inp of
                       Success (v1, n1, inp1, pos1) ->
                           fp `seq` (v1:v2, n1 + n2, inp2, fp)
                           where (v2, n2, inp2, pos2) = pListS p inp1
                                 fp = furthest_pos pos1 pos2
                       Failure pos -> ([], 0, inp, pos)
\end{code}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Lexing}

\advance\hsize-5em
\begin{tabular}{lcl}
$ascDigit$&$\rightarrow$&$0 \mid 1 \mid \ldots \mid 9$\cr
\end{tabular}

\begin{code}
lexc_ascDigit :: Parser Char Char
lexc_ascDigit = pPred isDigit
\end{code}

\begin{tabular}{lcl}
$\mathit{charesc}$&$\rightarrow$&$a \mid b \mid f \mid n \mid r \mid t
                                    \mid v \mid \backslash \mid "
                                    \mid \textrm{'} \mid \&$\cr
\end{tabular}

\begin{code}
lexs_charesc :: Parser Char String
lexs_charesc = wrap <$> lexc_charesc

lexc_charesc :: Parser Char Char
lexc_charesc = pSym 'a' <|  pSym 'b'  <|  pSym 'f' <|  pSym 'n'
           <|  pSym 'r' <|  pSym 't'  <|  pSym 'v' <|  pSym '\\'
           <|  pSym '"' <|  pSym '\'' <|  pSym '&'
\end{code}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{More lexing}

\advance\hsize-5em
\begin{tabular}{lcl}
$\mathit{varid}$&$\rightarrow$&$
  {\left(\mathit{small}\ \left\{ \mathit{small}
                            \mid \mathit{large}
                            \mid \mathit{digit}
                            \mid \textrm{'} \,\right\}\right)}
      _{<\mathit{reservedid}>}
  $\cr
$\mathit{conid}$&$\rightarrow$&$
  \mathit{large}\ \left\{ \mathit{small} \mid \mathit{large}
                     \mid \mathit{digit} \mid \textrm{'} \,\right\}
  $\cr
\end{tabular}

\begin{code}
id_body :: Parser Char Char
id_body = lexc_small <|  lexc_large <|  lexc_digit <|  pSym '\''
lexs_varid :: Parser Char String
lexs_varid = (:)
         <$> lexc_small
         <*> pList 0 id_body
         <!> lext_reservedid
lexs_conid :: Parser Char String
lexs_conid = (:)
         <$> lexc_large
         <*> pList 0 id_body
\end{code}

\begin{tabular}{lcl}
$\mathit{reservedop}$&$\rightarrow$&\verb!..! $\mid$ \verb!:! $\mid$
                                    \verb!::! $\mid$ \verb!=! $\mid$
                                    \verb!\! $\mid$ \verb!|! $\mid$
                                    \verb!<-! $\mid$ \verb!->! $\mid$
                                    \verb!@! $\mid$ \verb!~! $\mid$
                                    \verb!=>!\cr
\end{tabular}

\begin{code}
lext_reservedop :: Parser Char (Token, Position)
lext_reservedop = ReservedOp <&> lexs_reservedop
lexs_reservedop :: Parser Char String
lexs_reservedop = pSyms ".."  |> pSyms ":"   |> pSyms "::"
               |> pSyms "="   |> pSyms "\\"  |> pSyms "|"
               |> pSyms "<-"  |> pSyms "->"  |> pSyms "@"
               |> pSyms "~"   |> pSyms "=>"
\end{code}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Layout rule}

  2 stages:

  \begin{dinglist}[red]{\DingListSymbolA}
    \item Indent marking: Almost-code given in the report
    \item Applying the rule: Almost-code given in the report
  \end{dinglist}

  But! One of the conditions is parse-error(t)!

  Solution: Make an automaton for the (slightly simplified) Haskell grammar
  and step it each symbol

  Automaton constructed automatically from a CFG-like description

  Can also automatically create diagrams of the sub-automata, e.g.:

\includegraphics{diag2}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Fixity information}

  \begin{dinglist}[red]{\DingListSymbolA}
    \item Conceptually very simple
    \item Walk through the tree of blocks looking for fixity declarations, then
          apply them to the block and all child blocks
    \item Current code is a mess, but could be tidied up and made more efficient
  \end{dinglist}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Syntactic analysis}

Similar to lexical analysis, but need to optimise a number of cases, e.g.:

\vspace{1em}
Report:

\begin{tabular}{lcl}
$\mathit{exp}$&$\rightarrow$&$
    \mathit{exp0} :: \left[ \mathit{context}\ {=}{>} \right] \mathit{type}$\cr
&$\mid$&$\mathit{exp0}$\cr
\end{tabular}

\vspace{1em}
Optimised:

\begin{tabular}{lcl}
$\mathit{exp}$&$\rightarrow$&$
    \mathit{exp0} \left[ :: \left[ \mathit{context}\ {=}{>} \right]
                         \mathit{type} \right]$\cr
\end{tabular}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Pretty-printing}

Straightforward recursion over the abstract syntax tree

\begin{code}
s_op ">=" = "$\\ge$"
s_tyvar "a" = "$\\alpha$"
text = concatMap escape
escape '^' = "\\^{}"
pp_VarID v = "\\textit{" ++ text v ++ "\\/}"
pp_Idecl (Idecl_var var rhs, semi) = pp_Var var ++ pp_Rhs rhs semi
pp_Fixity Fixity_infixl = "\\HaskellToLaTeXkeyword{infixl}"
pp_Gd (Gd exp_i) = "\\HaskellToLaTeXguard{" ++ pp_Exp_i exp_i ++ "}"
\end{code}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{Some numbers}
  
\begin{tabular}{|l|l|}
\hline
\textbf{Size}&\textbf{CPU time taken for lexical analysis}\cr
\hline
355k&0m44.240s\cr
710k&1m29.760s\cr                                               
1420k&3m07.750s\cr                                              
\hline
\end{tabular}

\begin{tabular}{|l|l|}
\hline
\textbf{Using normal or optimised \hvarid{pList}}&\textbf{CPU time taken}\cr
\hline
Normal&10m39.840s\cr
Optimised&6m16.180s\cr
\hline
\end{tabular}

\begin{tabular}{|l|l|l|}
\hline
\textbf{Source size}&\textbf{CPU time taken}&\textbf{Processing speed (variable size heap)}\cr
\hline
44k&1m20.600s&551 bytes per second\cr
88k&2m43.620s&543 bytes per second\cr
176k&5m31.870s&535 bytes per second\cr
352k&11m14.850s&526 bytes per second\cr
704k&23m24.520s&505 bytes per second\cr
\hline
\end{tabular}

\begin{tabular}{|l|l|l|}
\hline
\textbf{Source size}&\textbf{CPU time taken}&\textbf{Processing speed (fixed size heap (300M))}\cr
\hline
44k&1m14.170s&598 bytes per second\cr
88k&2m25.540s&610 bytes per second\cr
176k&4m52.170s&608 bytes per second\cr
352k&9m42.200s&610 bytes per second\cr
704k&19m22.840s&611 bytes per second\cr
\hline
\end{tabular}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
  \slideheading{The path}
  
\begin{obeyspaces}
Original input:

\begin{tabular}{l}
\texttt{foo x = y + z}\cr
\texttt{    where y = 2 * x}\cr
\texttt{          z = x + 3}\cr
\end{tabular}

After indent marking:

\begin{tabular}{l}
\texttt{\{1\}foo x = y + z}\cr
\texttt{    <5>where \{11\}y = 2 * x}\cr
\texttt{          <11>z = x + 3}\cr
\end{tabular}

After layout:

\begin{tabular}{l}
\texttt{\{foo x = y + z}\cr
\texttt{    where \{y = 2 * x}\cr
\texttt{          ;z = x + 3}\cr
\texttt{\}\}}\cr
\end{tabular}

Parseable tokens:

\begin{tabular}{l}
\texttt{\{foox=y+zwhere\{y=2*x;z=x+3\}\}}\cr
\end{tabular}
\end{obeyspaces}

\vspace{0.5em}
Code:
\vspace{-0.5em}
\begin{code}
foo x = y + z
    where y = 2 * x
          z = x + 3
\end{code}
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\begin{slide}
\vspace*{\fill}
\slideheading{\Huge That's all folks!}
  
\vspace*{\fill}
\vspace*{\fill}
\end{slide}

\end{document}

